
import io

import numpy as np
import json
import asyncio
import logging
from typing import List,  Union,  Dict
from typing import AsyncGenerator, Optional, Union, List, Dict
#  Импортируем модели из models.py:
from .models import ProcessingSettings
import websockets
import httpx
from websockets.client import connect
from websockets.exceptions import ConnectionClosed
try:
    import sounddevice as sd
except ImportError:
    sd = None


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


        
class AudioClient:
    
    def __init__(
        self,
        base_url: str = "http://localhost:8000",
        timeout: float = 60.0,
        max_poll_attempts: int = 60,
        poll_interval: float = 1.0
    ):
        self.base_url = base_url.rstrip('/')
        self.timeout = timeout
        self.max_poll_attempts = max_poll_attempts
        self.poll_interval = poll_interval
        self.client = httpx.AsyncClient(timeout=timeout)

    async def __aenter__(self):
        await self.client.__aenter__()
        return self

    async def __aexit__(self, *args):
        await self.client.__aexit__(*args)


  # ---------------------------
  # Новая функция для работы с микрофоном
        
    async def async_microphone_stream(self, mic_index: int = None, sample_rate: int = 16000, channels: int = 1, blocksize: int = 1024):
        """
        Стримит аудио с микрофона в формате PCM 16-bit по websocket.
        
        ws_url: URL вебсокет-сервера, например "ws://localhost:8000/ws_transcribe"
        mic_index: индекс микрофона (None – использовать микрофон по умолчанию)
        """
        loop = asyncio.get_running_loop()
        queue = asyncio.Queue()
        ws_url = f"{self.base_url.replace('http', 'ws')}/ws_transcribe"
        def audio_callback(indata, frames, time, status):
            if status:
                print("Audio callback status:", status)
            # Если вы не задали dtype в InputStream, то данные будут float32.
            # Можно их конвертировать в int16:
            if indata.dtype != np.int16:
                audio_int16 = (indata * 32767).astype(np.int16)
            else:
                audio_int16 = indata
            # Преобразуем в байты и кладём в очередь
            data = audio_int16.tobytes()
            asyncio.run_coroutine_threadsafe(queue.put(data), loop)

        async with websockets.connect(ws_url, ping_interval=30, ping_timeout=60, close_timeout=1) as ws:
            # Отправляем начальные настройки (если это требуется сервером)
            settings = {"task": "transcribe", "transcription_settings": {"emotional": True, "language": "ru"}}
            await ws.send(json.dumps(settings))
            
            # Фоновая задача отправки аудио из очереди
            async def send_audio():
                try:
                    while True:
                        chunk = await queue.get()
                        await ws.send(chunk)
                except asyncio.CancelledError:
                    try:
                        await ws.send(b"")  # EOF-маркер
                    except:
                        pass
                    raise

            sender_task = asyncio.create_task(send_audio())
            
            try:
                # Открываем поток с явным указанием dtype='int16'
                with sd.InputStream(device=mic_index, samplerate=sample_rate, channels=channels, blocksize=blocksize, dtype='int16', callback=audio_callback):
                    print("Начата запись с микрофона (формат PCM 16-bit)")
                    while True:
                        try:
                            msg = await ws.recv()
                            yield json.loads(msg)
                        except websockets.ConnectionClosed:
                            break
            finally:
                sender_task.cancel()


# ---------------------------
# Вспомогательный метод для получения списка доступных микрофонов
    @staticmethod
    def list_microphones() -> List[Dict]:
        """
        Возвращает список устройств, способных к записи (микрофоны).
        Каждый элемент – словарь с описанием устройства.
        """
        devices = sd.query_devices()
        microphones = [dev for dev in devices if dev.get("max_input_channels", 0) > 0]
        return microphones

    async def process_audio(
        self,
        audio_input: Union[str, bytes, io.BytesIO],
        settings: Optional[ProcessingSettings] = None
    ) -> Union[dict, str]:
        settings = settings or ProcessingSettings()
        
        files, data = self._prepare_request(audio_input, settings)
        
        response = await self.client.post(
            f"{self.base_url}/process",
            data=data,
            files=files
        )
        response.raise_for_status()
        
        result = response.json()
        
        if not settings.async_process:
            return self._format_result(result, settings.format_output)
        
        return await self._handle_async_job(result.get('job_id'), settings.format_output)

    async def stream_audio(
        self,
        audio_input: Union[str, bytes],
        settings: Optional[ProcessingSettings] = None
    ) -> AsyncGenerator[dict, None]:
        settings = settings or ProcessingSettings()
        ws_url = f"{self.base_url.replace('http', 'ws')}/ws_transcribe"
        
        async with websockets.connect(
            ws_url,
            ping_interval=30,
            ping_timeout=60,
            close_timeout=1
        ) as ws:
            # Отправка настроек
            await ws.send(settings.json())
            
            async def sender():
                try:
                    if isinstance(audio_input, str):
                        with open(audio_input, "rb") as f:
                            while chunk := f.read(4000):
                                await ws.send(chunk)
                                await asyncio.sleep(0.001)
                    else:
                        chunk_size = 4000
                        for i in range(0, len(audio_input), chunk_size):
                            await ws.send(audio_input[i:i+chunk_size])
                            await asyncio.sleep(0.001)
                except ConnectionClosed:
                    return
                finally:
                    try:
                        await ws.send(b"")  # EOF-маркер
                        
                    except ConnectionClosed:
                        pass

            sender_task = asyncio.create_task(sender())
            
            try:
                while True:
                    try:
                        msg = await ws.recv()
                        yield json.loads(msg)
                    except ConnectionClosed as e:
                        #if e.code == 1000 or :  # Нормальное закрытие
                            break
                        #raise
            finally:
                await sender_task
                await ws.close()


    async def get_job_status(self, job_id: str) -> dict:
        response = await self.client.get(f"{self.base_url}/jobs/{job_id}")
        response.raise_for_status()
        return response.json()

    def _prepare_request(self, audio_input, settings):
        audio_bytes = self._get_audio_bytes(audio_input)
        return {
            "audio": ("audio.wav", audio_bytes, "audio/wav")
        }, {
            "mode": settings.mode,
            "task": settings.task,
            "format_output": settings.format_output,
            "async_process": str(settings.async_process).lower(),
            "transcription_settings": json.dumps(settings.transcription_settings)
        }

    def _get_audio_bytes(self, audio_input):
        if isinstance(audio_input, str):
            with open(audio_input, 'rb') as f:
                return f.read()
        elif isinstance(audio_input, bytes):
            return audio_input
        elif isinstance(audio_input, io.BytesIO):
            return audio_input.getvalue()
        raise ValueError("Unsupported input type")

    def _format_result(self, result, format_output):
        if format_output == "text":
            return self._extract_text(result)
        return result

    def _extract_text(self, result):
        if 'text' in result:
            return result['text']
        # Add logic to extract text from diarization results if needed
        return json.dumps(result)

    async def _handle_async_job(self, job_id, format_output):
        for _ in range(self.max_poll_attempts):
            await asyncio.sleep(self.poll_interval)
            status = await self.get_job_status(job_id)
            
            if status['status'] == 'completed':
                return self._format_result(status['result'], format_output)
            elif status['status'] == 'failed':
                raise ProcessingError(status['result'])
        
        raise TimeoutError("Job processing timed out")

class ProcessingError(Exception):
    pass